/*
 * Copyright (c) 2005-2010 KOM – Multimedia Communications Lab
 *
 * This file is part of PeerfactSim.KOM.
 *
 * PeerfactSim.KOM is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 *
 * PeerfactSim.KOM is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with PeerfactSim.KOM.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package de.tudarmstadt.maki.simonstrator.api.component.vehicular.roadnetwork;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.costs.PropertyImpactEstimator;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.costs.PropertyImpactEstimatorFactory;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.costs.VehicularPropertyImpactEstimator;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.properties.HazardProperty;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.properties.JamProperty;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.properties.RoadProperty;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.properties.vectoral.VectoralJamProperty;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.location.Location;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.api.EdgeController;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.roadnetwork.paths.VehiclePathTrackerFactory;

public class RoadNetworkEdge {
	private static final boolean ENABLE_SPEED_ADAPTATION = false;

	private String _edgeID;
	private double _angle;

	private double _originalMaxSpeed = 0;
	private double _maxSpeed = 0;

	private boolean _isUsableInit = false;
	private boolean _isUsable = false;

	private double _length = -1;

	private List<RoadNetworkLane> _lanes = new ArrayList<>();

	private List<RoadNetworkEdge> _accessibleEdges = new ArrayList<>();

	private List<RoadNetworkEdge> _incomingEdges = new ArrayList<>();

	private List<RoadProperty> _activeProperties = new ArrayList<>();

	private final EdgeController _edgeController;

	private List<RoadNetworkEdgeListener> _listeners = new ArrayList<>();

	public RoadNetworkEdge(SerializableRoadNetworkEdge pEdge, EdgeController pEdgeController) {
		_edgeController = pEdgeController;

		_edgeID = pEdge.getEdgeID();
		_angle = pEdge.getAngle();

		for (SerializableRoadNetworkLane lane : pEdge.getLanes()) {
			_lanes.add(new RoadNetworkLane(lane));
		}

		_originalMaxSpeed = pEdge.getMaxSpeed();
		_maxSpeed = _originalMaxSpeed;
	}

	public RoadNetworkEdge(String pEdgeID, double pAngle, EdgeController pEdgeController) {
		_edgeController = pEdgeController;

		_edgeID = pEdgeID;
		_angle = pAngle;
	}

	public void registerEdgeListener(RoadNetworkEdgeListener pListener) {
		_listeners.add(pListener);
	}

	public void addRoadProperty(RoadProperty pProperty) {
		_activeProperties.add(pProperty);

		adjustMaxSpeed();

		for (RoadNetworkEdgeListener roadNetworkEdgeListener : _listeners) {
			roadNetworkEdgeListener.edgeActivated(this, pProperty);
		}
	}

	private void adjustMaxSpeed() {
		if (ENABLE_SPEED_ADAPTATION) {
			double maxSpeed = getOriginalMaxSpeed();
			for (RoadProperty roadProperty : _activeProperties) {
				if (roadProperty instanceof JamProperty) {
					if (((JamProperty) roadProperty).getAverageSpeed() < maxSpeed) {
						maxSpeed = ((JamProperty) roadProperty).getAverageSpeed();
					}
				} else if (roadProperty instanceof HazardProperty) {
					if (0.001 < maxSpeed) {
						maxSpeed = 0.001;
					}
				}
			}

			changeEdgeMaxSpeed(maxSpeed);
		}
	}

	/**
	 * @return
	 */
	public double getCurrentSpeed() {
		double actualSpeed = -1;
		JamProperty jamProperty = getJamProperty();
		if (jamProperty != null) {
			actualSpeed = jamProperty.getAverageSpeed();
		} else {
			actualSpeed = getCorrespondingState(getOriginalMaxSpeed());
		}
		return actualSpeed;
	}

	public double getCurrentSpeedState() {
		return getCorrespondingState(getCurrentSpeed());
	}

	public JamProperty getJamProperty() {
		List<RoadProperty> activeProperties = getActiveProperties();
		for (RoadProperty roadProperty : activeProperties) {
			if (roadProperty instanceof JamProperty) {
				return (JamProperty) roadProperty;
			}
		}

		return null;
	}

	/**
	 * @param pExpectation
	 * @return
	 */
	public static double getCorrespondingState(double pExpectation) {
		return ((int) Math.round(pExpectation / VectoralJamProperty.SCALING)) * VectoralJamProperty.SCALING;
	}

	public void removeRoadProperty(RoadProperty pProperty) {
		_activeProperties.remove(pProperty);

		adjustMaxSpeed();

		for (RoadNetworkEdgeListener roadNetworkEdgeListener : _listeners) {
			roadNetworkEdgeListener.edgeDeactivated(this, pProperty);
		}
	}

	/**
	 * @return the activeProperties
	 */
	public List<RoadProperty> getActiveProperties() {
		return _activeProperties;
	}

	/**
	 * @param pActive
	 *            the active to set
	 */
	public void setActive(boolean pActive) {
		throw new UnsupportedOperationException();
	}

	/**
	 * @return the active
	 */
	public boolean isActive() {
		return _activeProperties.size() > 0;
	}

	public List<RoadNetworkLane> getLanes() {
		return _lanes;
	}

	public String getEdgeID() {
		return _edgeID;
	}

	public Location getEdgePosition() {
		return _edgeController.getEdgeMeanPosition(_edgeID);
	}

	public double getLastStepMeanSpeed() {
		return _edgeController.getLastStepMeanSpeed(_edgeID);
	}

	public double getAngle() {
		return _angle;
	}

	public double getOriginalMaxSpeed() {
		return _originalMaxSpeed;
	}

	public double getOriginalMaxSpeedState() {
		return RoadNetworkEdge.getCorrespondingState(_originalMaxSpeed);
	}

	public int getLaneAmount() {
		return _lanes.size();
	}

	public void addLane(RoadNetworkLane pLane) {
		_lanes.add(pLane);
	}

	public void setOriginalMaxSpeed(double pMaxSpeed) {
		if (pMaxSpeed > _originalMaxSpeed) {
			_originalMaxSpeed = pMaxSpeed;
			_maxSpeed = pMaxSpeed;
		}
	}

	public void setMaxSpeed(double pMaxSpeed) {
		_maxSpeed = pMaxSpeed;
	}

	public double getMaxSpeed() {
		return _maxSpeed;
	}

	public void addConnectedEdge(RoadNetworkEdge pEdge) {
		if (!_accessibleEdges.contains(pEdge)) {
			_accessibleEdges.add(pEdge);
			pEdge.addIncomingEdge(this);
		}
	}

	public void addIncomingEdge(RoadNetworkEdge pEdge) {
		if (!_incomingEdges.contains(pEdge)) {
			_incomingEdges.add(pEdge);
			pEdge.addConnectedEdge(this);
		}
	}

	public List<RoadNetworkEdge> getAccessibleEdges() {
		return Collections.unmodifiableList(_accessibleEdges);
	}

	public List<RoadNetworkEdge> getIncomingEdges() {
		return Collections.unmodifiableList(_incomingEdges);
	}

	public double getLength() {
		if (_length < 0) {
			_length = _edgeController.getEdgeLength(_edgeID);
		}
		return _length;
	}

	public String getDescription() {
		return "id: " + _edgeID + ", lanes: " + _lanes.size() + ", speed limit: " + _originalMaxSpeed;
	}

	@Override
	public String toString() {
		return _edgeID;
	}

	public boolean isInternal() {
		return _edgeID.startsWith(":");
	}

	public void changeEdgeMaxSpeed(double pMaxSpeed) {
		_maxSpeed = pMaxSpeed;
		_edgeController.setEdgeMaxSpeed(getEdgeID(), pMaxSpeed);
	}

	public boolean isUsable() {
		if (!_isUsableInit) {
			_isUsableInit = true;
			_isUsable = _edgeController.isEdgeUsable(getEdgeID());
		}
		return _isUsable;
	}

	public List<List<Location>> getRoadShape() {
		List<List<Location>> laneShapes = new ArrayList<>();

		for (RoadNetworkLane roadNetworkLane : _lanes) {
			List<Location> laneShape = _edgeController.getLaneShape(roadNetworkLane.getLaneID());
			laneShapes.add(laneShape);
		}

		return laneShapes;
	}

	/**
	 * Returns all road network edges, which guarantee that the a vehicle driving in
	 * the returned edge notifies any vehicle driving on the current edge.
	 * 
	 * @return
	 */
	public List<RoadNetworkEdge> getOpposingRoadNetworkEdges(double pCommunicationRange) {
		List<RoadNetworkEdge> edges = new ArrayList<>();

		for (RoadNetworkEdge roadNetworkEdge : RoadNetwork.CURRENT_ROAD_NETWORK.getUsableEdges()) {
			if (roadNetworkEdge != this) {

			}
		}

		return edges;
	}

	@Override
	public int hashCode() {
		return _edgeID.hashCode();
	}

	public boolean isJammed() {
		for (RoadProperty roadProperty : _activeProperties) {
			if (roadProperty instanceof JamProperty) {
				return ((JamProperty) roadProperty).isJammed();
			}
		}
		return false;
	}

	public double calculateStandardEdgeCosts() {
		return calculateStandardDrivingDuration();
	}

	public double calculateStandardDrivingDuration() {
		return VehiclePathTrackerFactory.getVehiclePathTracker().getTravelTime(this);
	}

	public double calculateEdgeCostIfKnown(RoadProperty pProperty) {
		double edgeCost = calculateStandardEdgeCosts();
		edgeCost += calculateAdditionalCostIfKnown(pProperty);
		return edgeCost;
	}

	public double calculateAdditionalCostIfKnown(RoadProperty pProperty) throws AssertionError {
		double additionalCost = 0;

		if (pProperty != null && pProperty.getDefaultProperty() != null
				&& !propertyEquals(pProperty, pProperty.getDefaultProperty())) {
			if (pProperty.getEdge().equals(this)) {
				PropertyImpactEstimator benefitEstimator = PropertyImpactEstimatorFactory
						.getPropertyBenefitEstimator();
				if (benefitEstimator instanceof VehicularPropertyImpactEstimator) {
					additionalCost = ((VehicularPropertyImpactEstimator) benefitEstimator)
							.calculateImpactIfKnown(pProperty);
				} else {
					throw new AssertionError("VehicularPropertyCostEstimator required!");
				}
			}
		}
		return additionalCost;
	}

	public double calculateEdgeCostIfUnknown(RoadProperty pProperty) {
		double edgeCost = calculateStandardEdgeCosts();
		edgeCost += calculateAdditionalCostIfUnknown(pProperty);
		return edgeCost;
	}

	public double calculateAdditionalCostIfUnknown(RoadProperty pProperty) throws AssertionError {
		double additionalCosts = 0;
		if (pProperty != null && !propertyEquals(pProperty, pProperty.getDefaultProperty())) {
			if (pProperty.getEdge().equals(this)) {
				PropertyImpactEstimator benefitEstimator = PropertyImpactEstimatorFactory
						.getPropertyBenefitEstimator();
				if (benefitEstimator instanceof VehicularPropertyImpactEstimator) {
					additionalCosts = ((VehicularPropertyImpactEstimator) benefitEstimator)
							.calculateImpactIfUnknown(pProperty);
				} else {
					throw new AssertionError("VehicularPropertyCostEstimator required!");
				}
			}
		}
		return additionalCosts;
	}

	private boolean propertyEquals(RoadProperty pFirst, RoadProperty pSecond) {
		return pFirst.getEdge().equals(pSecond.getEdge()) && pFirst.getValue().equals(pSecond.getValue())
				&& ((pFirst.getLocation() != null && pFirst.getLocation().equals(pSecond.getLocation()))
						|| (pFirst.getLocation() == null && pFirst.getLocation() == pSecond.getLocation()));

	}
}
